#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include "cpu.h"

#define TEST_EXPORT

#define MAX_INSTR 0x100000

#define _S_IREAD 0x0100
#define _S_IWRITE 0x0080

// end-detection
#define MAX_SILENT_FRAMES 1000

typedef unsigned char bool;
#define true 1
#define false 0


long filelength(int f)
{
    struct stat st;
    fstat(f, &st);
    return st.st_size;
}

int main(int argc, char **argv);
unsigned char readbyte(FILE *f);
unsigned short readword(FILE *f);

unsigned char prevChannelState[3][7];
unsigned char prevFilterState[4];
unsigned char tmpVals[7];

// freq lo, freq hi, pulse lo, pulse hi, ad, sr, wave&ctrl
unsigned char CHANNEL_REG_ORDER[7] = { 0x00, 0x01, 0x02, 0x03, 0x05, 0x06, 0x04 };
// cutoff lo, cutoff hi, reso&ctrl, mode&vol
unsigned char FILTER_REG_ORDER[4] = { 0x00, 0x01, 0x02, 0x03 };

unsigned char BITMASK_CHANNEL_WAVECTRL = 0x40;
unsigned char BITMASK_CHANNEL_SR = 0x20;
unsigned char BITMASK_CHANNEL_AD = 0x10;
unsigned char BITMASK_CHANNEL_PULSE_HI = 0x08;
unsigned char BITMASK_CHANNEL_PULSE_LO = 0x04;
unsigned char BITMASK_CHANNEL_FREQ_HI = 0x02;
unsigned char BITMASK_CHANNEL_FREQ_LO = 0x01;

unsigned char BITMASK_FILTER_MODEVOL = 0x08;
unsigned char BITMASK_FILTER_RESOCTRL = 0x04;
unsigned char BITMASK_FILTER_CUTOFF_HI = 0x02;
unsigned char BITMASK_FILTER_CUTOFF_LO = 0x01;

extern unsigned short pc;

int main(int argc, char **argv) {
	int subtune = 0;
	int seconds = 60 * 20;
	int instr = 0;
	int frames = 0;
	int firstframe = 0;
	int usage = 0;
	int totalframes = 0;
	unsigned loadend;
	unsigned loadpos;
	unsigned loadsize;
	unsigned loadaddress;
	unsigned initaddress;
	unsigned playaddress;
	unsigned dataoffset;
	unsigned nTunes;
	unsigned startTune;
	FILE *in;
	int out = -1;
	char *sidname = 0;
	char basename[512] = { 0 };
	char filename[512] = { 0 };

	// Scan arguments
	for (int c = 1; c < argc; c++) {
		if (argv[c][0] == '-') {
			switch (toupper(argv[c][1])) {
			case '?':
				usage = 1;
				break;

			case 'A':
				sscanf(&argv[c][2], "%u", &subtune);
				break;

			case 'F':
				sscanf(&argv[c][2], "%u", &firstframe);
				break;

			case 'T':
				sscanf(&argv[c][2], "%u", &seconds);
				break;
			}
		} else {
			if (!sidname)
				sidname = argv[c];
		}
	}

	// Usage display
	if ((argc < 2) || (usage)) {
		printf(
				"Usage: SIDDUMP <sidfile> [options]\n"
						"Warning: CPU emulation may be buggy/inaccurate, illegals support very limited\n\n"
						"Options:\n"
						"-a<value> Accumulator value on init (subtune number) default = 0\n"
						"-f<value> First frame to display, default 0\n"
						"-t<value> Playback time in seconds, default 60\n");
		return 1;
	}

	// Open SID file
	if (!sidname) {
		printf("Error: no SID file specified.\n");
		return 1;
	}

	printf("Loading %s\n", sidname);

	in = fopen(sidname, "rb");
	if (!in) {
		printf("Error: couldn't open SID file.\n");
		return 1;
	}

	// get the base filename of the sid file
	if (strlen(basename) == 0) {
		int c = strlen(sidname);
		while (sidname[c] != '.' && c > 0)
			c--;

		if (c != 0) {
			strncpy(basename, sidname, c);  // get the filename up to the .
			basename[c] = 0;
		} else
			strcpy(basename, sidname);  // otherwiuse just use the full filename
	}

	strcpy(filename, basename);
	strcat(filename, ".bin");
	printf("Writing %s\n", filename);

	out = open(filename, O_CREAT | O_TRUNC | O_RDWR, _S_IREAD | _S_IWRITE);
	if (out == -1) {
		printf("ERROR: failed to open %s\n", filename);
		fclose(in);
		return 1;
	}

#ifdef TEST_EXPORT
int channelOut[3] = { -1, -1, -1 };
int filterOut = -1;
char fn[1024];
strcpy(fn, filename);
strcat(fn, ".channel0.dat");
channelOut[0] = open(fn, O_CREAT | O_TRUNC | O_RDWR, _S_IREAD | _S_IWRITE);
strcpy(fn, filename);
strcat(fn, ".channel1.dat");
channelOut[1] = open(fn, O_CREAT | O_TRUNC | O_RDWR, _S_IREAD | _S_IWRITE);
strcpy(fn, filename);
strcat(fn, ".channel2.dat");
channelOut[2] = open(fn, O_CREAT | O_TRUNC | O_RDWR, _S_IREAD | _S_IWRITE);
strcpy(fn, filename);
strcat(fn, ".filter.dat");
filterOut = open(fn, O_CREAT | O_TRUNC | O_RDWR, _S_IREAD | _S_IWRITE);
#endif

	// Read interesting parts of the SID header
	fseek(in, 6, SEEK_SET);
	dataoffset = readword(in);
	loadaddress = readword(in);
	initaddress = readword(in);
	playaddress = readword(in);
	nTunes = readword(in);
	startTune = readword(in);

	fseek(in, dataoffset, SEEK_SET);

	if (loadaddress == 0)
		loadaddress = readbyte(in) | (readbyte(in) << 8);

	// Load the C64 data
	loadpos = ftell(in);
	fseek(in, 0, SEEK_END);
	loadend = ftell(in);
	fseek(in, loadpos, SEEK_SET);
	loadsize = loadend - loadpos;
	if (loadsize + loadaddress >= 0x10000) {
		printf("Error: SID data continues past end of C64 memory.\n");
		fclose(in);
		close(out);
		return 1;
	}
	fread(&mem[loadaddress], loadsize, 1, in);
	fclose(in);


	// Print info & run initroutine
	printf("Number of Tunes: %d\n", nTunes);
	printf("Start Tune     : %d\n", startTune);

//subtune = startTune - 1;
//subtune = 11;

	printf("Load address: $%04X Init address: $%04X Play address: $%04X\n", loadaddress, initaddress, playaddress);
	printf("Calling initroutine with subtune %d\n", subtune);
	mem[0x01] = 0x37;
	initcpu(initaddress, subtune, 0, 0);
	instr = 0;
	while (runcpu(true)) {
		instr++;
		if (instr > MAX_INSTR) {
			printf("Warning: CPU executed a high number of instructions in init, breaking\n");
			break;
		}
	}

	if (playaddress == 0) {
		printf("Warning: SID has play address 0, reading from interrupt vector instead\n");
		if ((mem[0x01] & 0x07) == 0x5)
			playaddress = mem[0xfffe] | (mem[0xffff] << 8);
		else
			playaddress = mem[0x314] | (mem[0x315] << 8);
		printf("New play address is $%04X\n", playaddress);
	}

	printf("Calling playroutine for %d frames, starting from frame %d\n", seconds * 50, firstframe);

	totalframes = firstframe + seconds * 50;

	//fprintf(outbinary, "SSND");
	//fputc((totalframes >> 8) & 0xFF, outbinary);
	//fputc(totalframes & 0xFF, outbinary);


	unsigned int nFramesStored = 0;
	unsigned int nSilentFrames = 0;	// trailing silent frames. used for end-detection.
	unsigned long fileEndPos = 0; 	// file pos after the last written non-empty frame. used to truncate the file after end-detection.

	bool channelGate[3] = { false, false, false };
	unsigned char filterVol = 0;

	// Data collection & display loop
	while (frames < totalframes) {

		memset(sidChannelMem, 0, MAX_MEM_WRITES * 3 * 7);
		memset(sidChannelMemWriteCounts, 0, MAX_MEM_WRITES * 3 * 7);
		memset(sidChannelMemMaxWriteCount, 0, 3);

		memset(sidFilterMem, 0, MAX_MEM_WRITES * 4);
		memset(sidFilterMemWriteCounts, 0, MAX_MEM_WRITES * 4);
		sidFilterMemMaxWriteCount = 0;

		// Run the playroutine
		instr = 0;
		initcpu(playaddress, 0, 0, 0);
		while (runcpu(false)) {
			instr++;
			if (instr > MAX_INSTR) {
				printf("Error: CPU executed abnormally high amount of instructions in playroutine, exiting\n");
				return 1;
			}
			// Test for jump into Kernal interrupt handler exit
			if ((mem[0x01] & 0x07) != 0x5 && (pc == 0xea31 || pc == 0xea81)) {
				break;
			}
		}

		if (frames == firstframe) {

			for (int channel = 0; channel < 3; ++channel) {
				unsigned char deltaFlags = 0x7f;
				int nVals = 0;
				for (int reg = 0; reg < 7; ++reg) {
					int r = CHANNEL_REG_ORDER[reg];
					int writeInd = sidChannelMemWriteCounts[channel][r] - 1;
					if (writeInd < 0) writeInd = 0;
					unsigned char val = sidChannelMem[writeInd][channel][r];
					prevChannelState[channel][r] = val;
					tmpVals[nVals++] = val;
					unsigned char bitMask = 0x01 << reg;

					if (bitMask == BITMASK_CHANNEL_WAVECTRL) channelGate[channel] = val & 0x01;
				}
				write(out, &deltaFlags, 1);
				if (nVals > 0) write(out, tmpVals, nVals);
			}

			unsigned char deltaFlags = 0xf;
			int nVals = 0;
			for (int reg = 0; reg < 4; ++reg) {
				int r = FILTER_REG_ORDER[reg];
				int writeInd = sidFilterMemWriteCounts[r] - 1;
				if (writeInd < 0) writeInd = 0;
				unsigned char val = sidFilterMem[writeInd][r];
				prevFilterState[r] = val;
				tmpVals[nVals++] = val;
				unsigned char bitMask = 0x01 << reg;

				if (bitMask == BITMASK_FILTER_MODEVOL) filterVol = val & 0x0f;
			}
			write(out, &deltaFlags, 1);
			if (nVals > 0) write(out, tmpVals, nVals);

			nFramesStored++;
		} else
		if (frames > firstframe) {

#ifdef TEST_EXPORT
unsigned char tmp[256];
unsigned char tmpLen;
#endif
			for (int channel = 0; channel < 3; ++channel) {
#ifdef TEST_EXPORT
tmpLen = 0;
#endif
				if (sidChannelMemMaxWriteCount[channel] == 0) sidChannelMemMaxWriteCount[channel] = 1;
				for (int writeInd = 0; writeInd < sidChannelMemMaxWriteCount[channel]; ++writeInd) {
					unsigned char deltaFlags = 0x00;
					if (writeInd < sidChannelMemMaxWriteCount[channel] - 1) deltaFlags |= 0x80;
					int nVals = 0;
					for (int reg = 0; reg < 7; ++reg) {
						int r = CHANNEL_REG_ORDER[reg];
						int writeCount = sidChannelMemWriteCounts[channel][r];
						if (writeCount <= writeInd) continue;
						unsigned char val = sidChannelMem[writeInd][channel][r];
						unsigned char prevVal = prevChannelState[channel][r];
						if (prevVal == val) continue;
						prevChannelState[channel][r] = val;
						tmpVals[nVals++] = val;
						unsigned char bitMask = 0x01 << reg;
						deltaFlags |= bitMask;

						if (bitMask == BITMASK_CHANNEL_WAVECTRL) channelGate[channel] = val & 0x01;
					}
					write(out, &deltaFlags, 1);
					if (nVals > 0) write(out, tmpVals, nVals);

#ifdef TEST_EXPORT
tmp[tmpLen++] = deltaFlags;
for (int i = 0; i < nVals; ++i) tmp[tmpLen++] = tmpVals[i];
				}
write(channelOut[channel], &tmpLen, 1);
write(channelOut[channel], tmp, tmpLen);
#endif
			}

#ifdef TEST_EXPORT
tmpLen = 0;
#endif
			if (sidFilterMemMaxWriteCount == 0) sidFilterMemMaxWriteCount = 1;
			for (int writeInd = 0; writeInd < sidFilterMemMaxWriteCount; ++writeInd) {
				unsigned char deltaFlags = 0x00;
				if (writeInd < sidFilterMemMaxWriteCount - 1) deltaFlags |= 0x10;
				int nVals = 0;
				for (int reg = 0; reg < 4; ++reg) {
					int r = FILTER_REG_ORDER[reg];
					int writeCount = sidFilterMemWriteCounts[r];
					if (writeCount <= writeInd) continue;
					unsigned char val = sidFilterMem[writeInd][r];
					unsigned char prevVal = prevFilterState[r];
					if (prevVal == val) continue;
					prevFilterState[r] = val;
					tmpVals[nVals++] = val;
					unsigned char bitMask = 0x01 << reg;
					deltaFlags |= bitMask;

					if (bitMask == BITMASK_FILTER_MODEVOL) filterVol = val & 0x0f;
				}
				write(out, &deltaFlags, 1);
				if (nVals > 0) write(out, tmpVals, nVals);
#ifdef TEST_EXPORT
tmp[tmpLen++] = deltaFlags;
for (int i = 0; i < nVals; ++i) tmp[tmpLen++] = tmpVals[i];
write(filterOut, &tmpLen, 1);
write(filterOut, tmp, tmpLen);
#endif
			}


			bool isChannelGateOn = channelGate[0] | channelGate[1] | channelGate[2];

			bool isSilent = (isChannelGateOn == false) || (filterVol == 0);

			if (isSilent) {
				nSilentFrames++;
				if (nSilentFrames > MAX_SILENT_FRAMES) {
					nFramesStored -= MAX_SILENT_FRAMES;
					double sec = nFramesStored / 50.0;
					int min = (int) (sec / 60);
					sec -= min * 60;
					printf("END OF TUNE DETECTED AFTER %d frames (%d:%f)\n", nFramesStored, min, sec);
					break;
				}
			} else {
				nSilentFrames = 0;
				fileEndPos = filelength(out);
			}

			nFramesStored++;
		}

		frames++;
	}

//	chsize(out, fileEndPos);
	ftruncate(out, fileEndPos);
	close(out);

#ifdef TEST_EXPORT
close(channelOut[0]);
close(channelOut[1]);
close(channelOut[2]);
close(filterOut);
#endif
	return 0;
}

unsigned char readbyte(FILE *f) {
	unsigned char res;

	fread(&res, 1, 1, f);
	return res;
}

unsigned short readword(FILE *f) {
	unsigned char res[2];

	fread(&res, 2, 1, f);
	return (res[0] << 8) | res[1];
}
